from typing import Callable

from celestia.types.common_types import Namespace, Base64
from celestia.types.header import ExtendedHeader
from celestia.types.share import ExtendedDataSquare, NamespaceData, SampleCoords, GetRangeResult
from ._RPC import Wrapper


class ShareClient(Wrapper):
    """ Client for interacting with Celestia's Share API."""

    async def get_eds(self, height: int, deserializer: Callable | None = None) -> ExtendedDataSquare:
        """ Gets the full EDS identified by the given extended header.

        Args:
            height (int): The block height.
            deserializer (Callable | None): Custom deserializer. Defaults to :meth:`~celestia.types.share.ExtendedDataSquare.deserializer`.

        Returns:
            ExtendedDataSquare: The retrieved EDS object.
        """

        deserializer = deserializer if deserializer is not None else ExtendedDataSquare.deserializer

        return await self._rpc.call("share.GetEDS", (height,), deserializer)

    async def get_namespace_data(self, height: int, namespace: Namespace,
                                 deserializer: Callable | None = None) -> list[NamespaceData]:
        """ Gets all shares from an EDS within the given namespace. Shares are returned
        in a row-by-row order if the namespace spans multiple rows.

        Args:
            height (int): The block height.
            namespace (Namespace): The namespace identifier.
            deserializer (Callable | None): Custom deserializer. Defaults to None.

        Returns:
            list[NamespaceData]: A list of NamespaceData objects.
        """

        def deserializer_(result):
            if result is not None:
                return [NamespaceData(**data) for data in result]

        deserializer = deserializer if deserializer is not None else deserializer_

        return await self._rpc.call("share.GetNamespaceData", (height, Namespace(namespace)), deserializer)

    async def get_range(self, height: int, start: int, end: int,
                        deserializer: Callable | None = None) -> GetRangeResult:
        """ Gets a list of shares and their corresponding proof.

        Args:
            height (int): The block height.
            start (int): The starting index.
            end (int): The ending index.
            deserializer (Callable | None): Custom deserializer. Defaults to :meth:`~celestia.types.share.GetRangeResult.deserializer`.

        Returns:
            GetRangeResult: The retrieved range result containing shares and proof.
        """

        deserializer = deserializer if deserializer is not None else GetRangeResult.deserializer

        return await self._rpc.call("share.GetRange", (height, start, end), deserializer)

    async def get_samples(self, header: ExtendedHeader, indices: list[SampleCoords]) -> list[Base64]:
        """ Gets sample for given indices.

        Args:
            header (ExtendedHeader): The extended header.
            indices (list[SampleCoords]): A list of sample coordinates.

        Returns:
            list[Base64]: A list of retrieved samples.
        """
        return await self._rpc.call("share.GetSamples", (header, indices,))

    async def get_share(self, height: int, row: int, col: int) -> Base64:
        """ Gets a Share by coordinates in EDS.

        Args:
            height (int): The block height.
            row (int): The row index.
            col (int): The column index.

        Returns:
            Base64: The retrieved share.
        """
        return await self._rpc.call("share.GetShare", (height, row, col,))

    async def get_available(self, height: int) -> bool:
        """ Subjectively validates if Shares committed to the given ExtendedHeader are available on the Network.

        Args:
            height (int): The block height.

        Returns:
            bool: True if shares are available, False otherwise.
        """
        return await self._rpc.call("share.SharesAvailable", (height,))



import logging
import os.path
import subprocess
from time import sleep

from tests.docker import Container

TESTNET_PATH = os.path.dirname(os.path.abspath(__file__))


def start_testnet() -> bool:
    args = ['docker', 'compose', '-f', f'{TESTNET_PATH}/testnet/docker-compose.yml', 'up', '-d']
    proc = subprocess.run(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    if proc.returncode == 0:
        return True
    logging.error(proc.stderr.decode('utf8'))
    return False


def stop_testnet() -> bool:
    args = ['docker', 'compose', '-f', f'{TESTNET_PATH}/testnet/docker-compose.yml', 'down']
    proc = subprocess.run(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    if proc.returncode == 0:
        return True
    logging.error(proc.stderr.decode('utf8'))
    return False


def get_auth_token(container: Container) -> str | None:
    args = ['docker', 'exec', '-i', container.id,
            'celestia', container.image, 'auth', 'admin', '--p2p.network', 'private']
    cnt = 10
    while cnt:
        proc = subprocess.run(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        if proc.returncode == 0:
            token = proc.stdout.decode('utf8').strip()
            return token
        sleep(1)
        cnt -= 1
        if not cnt:
            logging.error(proc.stderr.decode('utf8'))



import asyncio
from time import sleep

import pytest

from celestia.node_api import Client
from tests.docker import Containers
from tests.utils import start_testnet, stop_testnet, get_auth_token


@pytest.fixture(scope="session")
def containers():
    cnt = 10
    need_shutdown = False
    containers = Containers('testnet')
    while cnt:
        if containers:
            break
        if not need_shutdown:
            start_testnet()
            need_shutdown = True
        cnt -= 1
        sleep(10 - cnt)
        containers = Containers('testnet')
    else:
        RuntimeError("Cannot start testnet")
    yield containers
    if need_shutdown:
        stop_testnet()

@pytest.fixture(scope="session")
def ready_nodes():
    yield dict()

@pytest.fixture
def node_provider(containers, ready_nodes):
    #
    async def node_provider_(name):
        if name in ready_nodes:
            return ready_nodes[name]
        elif node := containers.get_by_name_first(name):
            auth_token = get_auth_token(node)
            cnt = 10
            while cnt:
                try:
                    async with Client(port=node.port['26658/tcp']).connect(auth_token) as api:
                        balance = await api.state.balance()
                        if balance.amount:
                            ready_nodes[name] = node, auth_token
                            return ready_nodes[name]
                except:
                    pass
                cnt -= 1
                await asyncio.sleep(10 - cnt)
            else:
                raise RuntimeError(f"Node '{name}' not ready")
        else:
            raise RuntimeError(f"Node '{name}' not found")

    return lambda name: node_provider_(name)


@pytest.fixture(scope='session')
def light_address():
    yield ('celestia1ll9pjlvy8cg7ux3pr98sc96nlpwgzt48j2mjwz',)


@pytest.fixture(scope='session')
def bridge_addresses():
    yield ('celestia1t52q7uqgnjfzdh3wx5m5phvma3umrq8k6tq2p9',
           'celestia16ws8cxx9ykl4598qgshvt36mejpkeyvzayndth',
           'celestia10yeexpgcpx88qru4ca63frhw3jqua4qw8swxy0')


@pytest.fixture(scope='session')
def validator_addresses():
    yield ('celestiavaloper1tzkpek429yxtvrshqh5yvqhvq4ydu3pjrshjhh',
           'celestiavaloper1uqmt6u5zwzucxjkg7pd30qw8lc6l4c8xxv9288',
           'celestiavaloper12crcjleegs25gp8wdx3nwn2m9kvfdmc34apd28')



from collections.abc import AsyncIterator
from functools import wraps
from typing import Callable, Unpack

from celestia._celestia import types  # noqa

from celestia.types.blob import SubmitBlobResult, Proof, CommitmentProof, SubscriptionBlobResult
from celestia.types.common_types import Blob, Namespace, TxConfig, Commitment
from ._RPC import Wrapper


def handle_blob_error(func):
    """ Decorator to handle blob-related errors."""

    @wraps(func)
    async def wrapper(*args, **kwargs):
        try:
            return await func(*args, **kwargs)
        except ConnectionError as e:
            if 'blob: not found' in e.args[1].body['message'].lower():
                return None
            raise

    return wrapper


class BlobClient(Wrapper):
    """ Client for interacting with Celestia's Blob API."""

    @handle_blob_error
    async def get(self, height: int, namespace: Namespace, commitment: Commitment,
                  deserializer: Callable | None = None) -> Blob | None:
        """ Retrieves the blob by commitment under the given namespace and height.

        Args:
            height (int): The block height.
            namespace (Namespace): The namespace of the blob.
            commitment (Commitment): The commitment of the blob.
            deserializer (Callable | None): Custom deserializer. Defaults to :meth:`~celestia.types.common_types.Blob.deserializer`.

        Returns:
            Blob | None: The retrieved blob, or None if not found.
        """

        deserializer = deserializer if deserializer is not None else Blob.deserializer

        return await self._rpc.call("blob.Get", (height, Namespace(namespace), Commitment(commitment)), deserializer)

    async def get_all(self, height: int, namespace: Namespace, *namespaces: Namespace,
                      deserializer: Callable | None = None) -> list[Blob] | None:
        """ Returns all blobs under the given namespaces at the given height. If all blobs were
        found without any errors, the user will receive a list of blobs. If the BlobService couldn't
        find any blobs under the requested namespaces, the user will receive an empty list of blobs
        along with an empty error. If some of the requested namespaces were not found, the user will receive
        all the found blobs and an empty error. If there were internal errors during some of the requests, the
        user will receive all found blobs along with a combined error message. All blobs will preserve the order
        of the namespaces that were requested.

        Args:
            height (int): The block height.
            namespace (Namespace): The primary namespace of the blobs.
            namespaces (Namespace): Additional namespaces to query for blobs.
            deserializer (Callable | None): Custom deserializer. Defaults to None.

        Returns:
            list[Blob] | None: The list, of blobs or None if not found.
        """

        def deserializer_(result):
            if result is not None:
                return [Blob(**kwargs) for kwargs in result]

        deserializer = deserializer if deserializer is not None else deserializer_
        namespaces = tuple(Namespace(namespace) for namespace in (namespace, *namespaces))

        return await self._rpc.call("blob.GetAll", (height, namespaces), deserializer)

    async def submit(self, blob: Blob, *blobs: Blob, deserializer: Callable | None = None,
                     **options: Unpack[TxConfig]) -> SubmitBlobResult:
        """ Sends Blobs and reports the height in which they were included. Allows sending
        multiple Blobs atomically synchronously. Uses default wallet registered on the Node.

        Args:
            blob (Blob): The main blob to submit.
            blobs (Blob): Additional blobs to submit.
            deserializer (Callable | None): Custom deserializer. Defaults to None.
            options (TxConfig): Additional configuration options.

        Returns:
            SubmitBlobResult: The result of the submission, including the height.
        """

        def deserializer_(height):
            if height is not None:
                return SubmitBlobResult(height, tuple(blob.commitment for blob in blobs))

        deserializer = deserializer if deserializer is not None else deserializer_
        blobs = tuple(types.normalize_blob(blob) if blob.commitment is None else blob for blob in (blob, *blobs))

        return await self._rpc.call("blob.Submit", (blobs, options), deserializer)

    @handle_blob_error
    async def get_commitment_proof(self, height: int, namespace: Namespace, commitment: Commitment,
                                   deserializer: Callable | None = None) -> CommitmentProof | None:
        """ Generates a commitment proof for a share commitment.

        Args:
            height (int): The block height.
            namespace (Namespace): The namespace of the commitment.
            commitment (Commitment): The commitment to generate proof for.
            deserializer (Callable | None): Custom deserializer. Defaults to :meth:`~celestia.types.blob.CommitmentProof.deserializer`.

        Returns:
            CommitmentProof | None: The commitment proof, or None if not found.
        """

        deserializer = deserializer if deserializer is not None else CommitmentProof.deserializer

        return await self._rpc.call("blob.GetCommitmentProof", (height, Namespace(namespace), Commitment(commitment)),
                                    deserializer)

    @handle_blob_error
    async def get_proof(self, height: int, namespace: Namespace, commitment: Commitment,
                        deserializer: Callable | None = None) -> list[Proof] | None:
        """ Retrieves proofs in the given namespaces at the given height by commitment.

        Args:
            height (int): The block height.
            namespace (Namespace): The namespace of the proof.
            commitment (Commitment): The commitment to generate the proof for.
            deserializer (Callable | None): Custom deserializer. Defaults to :meth:`~celestia.types.blob.Proof.deserializer`.

        Returns:
            list[Proof] | None: A list of proofs, or None if not found.
        """

        deserializer = deserializer if deserializer is not None else Proof.deserializer

        return await self._rpc.call("blob.GetProof", (height, Namespace(namespace), Commitment(commitment)),
                                    deserializer)

    async def included(self, height: int, namespace: Namespace, proof: Proof, commitment: Commitment) -> bool:
        """ Checks whether a blob's given commitment(Merkle subtree root) is
        included at given height and under the namespace.

        Args:
            height (int): The block height.
            namespace (Namespace): The namespace of the blob.
            proof (Proof): The proof to check.
            commitment (Commitment): The commitment to check inclusion for.

        Returns:
            bool: True if included, False otherwise.
        """

        return await self._rpc.call("blob.Included", (height, Namespace(namespace), proof, Commitment(commitment)))

    async def subscribe(self, namespace: Namespace,
                        deserializer: Callable | None = None) -> AsyncIterator[SubscriptionBlobResult | None]:
        """ Subscribe to published blobs from the given namespace as they are included.

        Args:
            namespace (Namespace): The namespace to subscribe to.
            deserializer (Callable | None): Custom deserializer. Defaults to None.

        Yields:
            SubscriptionBlobResult: A result with the blob information.

        """

        def deserializer_(result):
            height = result['Height']
            blobs = result.get('Blobs')
            if blobs is not None:
                return SubscriptionBlobResult(height, tuple(Blob(**kwargs) for kwargs in blobs))

        deserializer = deserializer if deserializer is not None else deserializer_

        async for subs_blob_result in self._rpc.iter("blob.Subscribe", (Namespace(namespace),), deserializer):
            if subs_blob_result is not None:
                yield subs_blob_result



from dataclasses import dataclass
from enum import Enum
from typing import Any


@dataclass
class ResourceManagerStat:
    """ Represents the statistics of a resource manager, including system, transient,
    services, protocols, and peers.

    Attributes:
        system (dict): System data.
        transient (dict): Transient data.
        services (dict): Service data.
        protocols (dict): Protocol data.
        peers (dict): Peer data.
    """
    system: dict[str, Any]
    transient: dict[str, Any]
    services: dict[str, Any]
    protocols: dict[str, Any]
    peers: dict[str, Any]

    def __init__(self, System, Transient, Services, Protocols, Peers):
        self.system = System
        self.transient = Transient
        self.services = Services
        self.protocols = Protocols
        self.peers = Peers

    @staticmethod
    def deserializer(result: dict) -> 'ResourceManagerStat':
        """ Deserialize a dictionary into a ResourceManagerStat instance.

        Args:
            result (dict): The dictionary representation of a ResourceManagerStat.

        Returns:
            ResourceManagerStat: A deserialized ResourceManagerStat object.
        """
        if result is not None:
            return ResourceManagerStat(**result)


@dataclass
class BandwidthStats:
    """ Represents the statistics related to bandwidth, including total inbound/outbound
    traffic and rates for both directions.

    Attributes:
        total_in (int): Total inbound bandwidth.
        total_out (int): Total outbound bandwidth.
        rate_in (float): Inbound traffic rate.
        rate_out (float): Outbound traffic rate.
    """
    total_in: int
    total_out: int
    rate_in: float
    rate_out: float

    def __init__(self, TotalIn, TotalOut, RateIn, RateOut):
        self.total_in = TotalIn
        self.total_out = TotalOut
        self.rate_in = RateIn
        self.rate_out = RateOut

    @staticmethod
    def deserializer(result: dict) -> 'BandwidthStats':
        """ Deserialize a dictionary into a BandwidthStats instance.

        Args:
            result (dict):  The dictionary representation of a BandwidthStats.

        Returns:
            BandwidthStats: A deserialized BandwidthStats object.
        """
        if result is not None:
            return BandwidthStats(**result)


@dataclass
class AddrInfo:
    """ Represents the address information with an identifier and associated addresses.

    Attributes:
        id (str): The unique identifier for the address.
        addrs (list): A list of addresses associated with the identifier.
    """
    id: str
    addrs: list[str]

    def __init__(self, ID, Addrs):
        self.id = ID
        self.addrs = Addrs

    @staticmethod
    def deserializer(result: dict) -> 'AddrInfo':
        """ Deserialize a dictionary into an AddrInfo instance.

        Args:
           result (dict): The dictionary representation of a AddrInfo.

        Returns:
           AddrInfo: A deserialized AddrInfo object.
        """
        if result is not None:
            return AddrInfo(**result)


class Connectedness(Enum):
    """ Enum representing the connection status.

    Attributes:
        NOT_CONNECTED: Represents a disconnected state.
        CONNECTED: Represents a connected state.
    """
    NOT_CONNECTED = 0
    CONNECTED = 1


class Reachability(Enum):
    """ Enum representing the reachability state of an address.

    Attributes:
       Unknown: Unknown reachability state.
       Public: The address is publicly reachable.
       Private: The address is privately reachable.
    """
    Unknown = 0
    Public = 1
    Private = 2



from dataclasses import asdict

import pytest

from celestia.node_api import Client
from celestia.node_api.p2p import Connectedness, Reachability


@pytest.mark.asyncio
async def test_p2p(node_provider):
    bridge1, auth_token1 = await node_provider('bridge-0')
    bridge2, auth_token2 = await node_provider('bridge-1')
    bridge3, auth_token3 = await node_provider('bridge-2')
    client1 = Client(port=bridge1.port['26658/tcp'])
    client2 = Client(port=bridge2.port['26658/tcp'])
    client3 = Client(port=bridge3.port['26658/tcp'])

    async with client1.connect(auth_token1) as api:
        info1 = await api.p2p.info()

    async with client2.connect(auth_token2) as api:
        info2 = await api.p2p.info()

    async with client3.connect(auth_token3) as api:
        info3 = await api.p2p.info()

        assert Connectedness.CONNECTED.value == await api.p2p.connectedness(info1.id)
        assert Connectedness.CONNECTED.value == await api.p2p.connectedness(info2.id)

        await api.p2p.close_peer(info1.id)
        assert Connectedness.NOT_CONNECTED.value == await api.p2p.connectedness(info1.id)
        assert info1.id not in await api.p2p.peers()
        await api.p2p.connect(info1)
        assert Connectedness.CONNECTED.value == await api.p2p.connectedness(info1.id)
        assert info1.id in await api.p2p.peers()

        await api.p2p.block_peer(info2.id)
        assert Connectedness.NOT_CONNECTED.value == await api.p2p.connectedness(info2.id)
        assert info2.id in await api.p2p.list_blocked_peers()
        await api.p2p.unblock_peer(info2.id)
        await api.p2p.connect(info2)
        assert Connectedness.CONNECTED.value == await api.p2p.connectedness(info2.id)
        assert info2.id not in await api.p2p.list_blocked_peers()

        one_peer_stats = await api.p2p.bandwidth_for_peer(info1.id)
        assert not all(value == 0 for value in asdict(one_peer_stats).values())
        bad_peer_stats = await api.p2p.bandwidth_for_peer('12D3KooWEhKP6kFF3Ptz14PeJkBXT4RNF8Dbc6LVa4gMbBVeajkQ')
        assert all(value == 0 for value in asdict(bad_peer_stats).values())
        all_peers_stats = await api.p2p.bandwidth_stats()
        assert not all(value == 0 for value in asdict(all_peers_stats).values())

        res_man_stat = await api.p2p.resource_state()

        for keys in res_man_stat.protocols.keys():
            protocol_stats = await api.p2p.bandwidth_for_protocol(keys)
            assert not all(value == 0 for value in asdict(protocol_stats).values())

        bad_protocol_stats = await api.p2p.bandwidth_for_protocol('/qwe/asd/zxc')
        assert all(value == 0 for value in asdict(bad_protocol_stats).values())

        assert not await api.p2p.is_protected(info1.id, 'Test tag')
        await api.p2p.protect(info1.id, 'Test tag')
        assert await api.p2p.is_protected(info1.id, 'Test tag')
        assert not await api.p2p.is_protected(info1.id, 'Bad test tag')
        await api.p2p.unprotect(info1.id, 'Test tag')
        assert not await api.p2p.is_protected(info1.id, 'Test tag')

        assert Reachability.Unknown.value == await api.p2p.nat_status()

        topic = await api.p2p.pub_sub_topics()
        assert len(await api.p2p.pub_sub_peers(topic[0])) != 0
        assert await api.p2p.pub_sub_peers('Bad test topic') is None

        with pytest.raises(ValueError):
            await api.p2p.connect(info3)
        with pytest.raises(ValueError):
            await api.p2p.block_peer(info2.id)
            await api.p2p.connect(info2)
        await api.p2p.unblock_peer(info2.id)



from dataclasses import dataclass

from celestia.types.blob import RowProof, Proof
from celestia.types.common_types import Base64, Namespace


@dataclass
class SampleCoords:
    """ A class representing coordinates for a sample, specifically the row and column.

    Attributes:
        row (int): The row index of the sample.
        col (int): The column index of the sample.
    """
    row: int
    col: int


@dataclass
class ShareProof:
    """ A class representing a share proof, which consists of a namespace ID,
    namespace version, row proof, data, and share proofs.

    Attributes:
        namespace_id (Namespace): The namespace identifier.
        namespace_version (int): The version of the namespace.
        row_proof (RowProof): The proof related to the row.
        data (tuple[Base64, ...]): The data associated with the share proof.
        share_proofs (tuple[Proof, ...]): Additional share proofs.
    """
    namespace_id: Namespace
    namespace_version: int
    row_proof: RowProof
    data: tuple[Base64, ...]
    share_proofs: tuple[Proof, ...]

    def __init__(self, namespace_id: Namespace, namespace_version: int,
                 row_proof: dict, data: tuple[Base64, ...],
                 share_proofs: list[dict]):
        self.namespace_id = Namespace.ensure_type(namespace_id)
        self.namespace_version = int(namespace_version)
        self.row_proof = RowProof(**row_proof)
        self.data = tuple(data_unit for data_unit in data)
        self.share_proofs = tuple(Proof(**share_proof) for share_proof in share_proofs)


@dataclass
class GetRangeResult:
    """
    A class representing the result of a range retrieval, including shares and proof.

    Attributes:
        shares (tuple[Base64, ...]): The shares related to the range.
        proof (ShareProof): The proof associated with the range retrieval.
    """
    shares: tuple[Base64, ...]
    proof: ShareProof

    def __init__(self, Shares: list[Base64], Proof: dict):
        self.shares = tuple(share for share in Shares)
        self.proof = ShareProof(**Proof)

    @staticmethod
    def deserializer(result: dict) -> 'GetRangeResult':
        """ Deserialize a result dictionary into a GetRangeResult object.

        Args:
            result (dict): The dictionary representation of a GetRangeResult.

        Returns:
            GetRangeResult: The deserialized GetRangeResult object.
        """
        if result is not None:
            return GetRangeResult(**result)


@dataclass
class ExtendedDataSquare:
    """ A class representing an extended data square, including the data square and codec.

    Attributes:
        data_square (tuple[Base64, ...]): The data square.
        codec (str): The codec used for the data.
    """
    data_square: tuple[Base64, ...]
    codec: str

    @staticmethod
    def deserializer(result: dict) -> 'ExtendedDataSquare':
        """ Deserialize a result dictionary into an ExtendedDataSquare object.

        Args:
            result (dict): The dictionary representation of a ExtendedDataSquare.

        Returns:
            ExtendedDataSquare: The deserialized ExtendedDataSquare object.
        """
        if result is not None:
            return ExtendedDataSquare(**result)


@dataclass
class NamespaceData:
    """ A class representing namespace data, consisting of shares and proof.

    Attributes:
        shares (tuple[Base64, ...]): The shares related to the namespace.
        proof (Proof): The proof associated with the namespace data.
    """
    shares: tuple[Base64, ...]
    proof: Proof

    def __init__(self, shares: list[Base64], proof: dict):
        self.shares = tuple(share for share in shares)
        self.proof = Proof(**proof)

    @staticmethod
    def deserializer(result: dict) -> 'NamespaceData':
        """ Deserialize a result dictionary into a NamespaceData object.

        Args:
            result (dict): The dictionary representation of a NamespaceData.

        Returns:
            NamespaceData: The deserialized NamespaceData object.
        """
        if result is not None:
            return NamespaceData(**result)



from dataclasses import dataclass


@dataclass
class Worker:
    job_type: str
    current: int
    from_: int
    to: int


@dataclass
class SamplingStats:
    head_of_sampled_chain: int
    head_of_catchup: int
    network_head_height: int
    concurrency: int
    catch_up_done: bool
    is_running: bool
    workers: tuple[Worker, ...] = None

    def __init__(self, head_of_sampled_chain, head_of_catchup, network_head_height, concurrency, catch_up_done,
                 is_running, workers=None):
        self.head_of_sampled_chain = head_of_sampled_chain
        self.head_of_catchup = head_of_catchup
        self.network_head_height = network_head_height
        self.concurrency = concurrency
        self.catch_up_done = catch_up_done
        self.is_running = is_running
        self.workers = tuple(Worker(**worker) for worker in workers) if workers is not None else None

    @staticmethod
    def deserializer(result):
        if result is not None:
            return SamplingStats(**result)



import asyncio

import pytest

from celestia.node_api import Client
from celestia.node_api.blob import Blob


@pytest.mark.asyncio
async def test_account_address(node_provider, bridge_addresses):
    bridge, auth_token = await node_provider('bridge-0')
    client = Client(port=bridge.port['26658/tcp'])
    async with client.connect(auth_token) as api:
        address = await api.state.account_address()
        assert address in bridge_addresses


@pytest.mark.asyncio
async def test_account_balance(node_provider):
    bridge, auth_token = await node_provider('bridge-0')
    client = Client(port=bridge.port['26658/tcp'])
    async with client.connect(auth_token) as api:
        balance = await api.state.balance()
        assert balance.amount > 100000000000000
        assert balance.denom == 'utia'
        address = await api.state.account_address()
        address_balance = await api.state.balance_for_address(address)
        assert address_balance == balance


@pytest.mark.asyncio
async def test_transfer(node_provider):
    bridge1, auth_token1 = await node_provider('bridge-0')
    client1 = Client(port=bridge1.port['26658/tcp'])
    bridge2, auth_token2 = await node_provider('bridge-1')
    client2 = Client(port=bridge2.port['26658/tcp'])
    bridge3, auth_token3 = await node_provider('bridge-2')
    client3 = Client(port=bridge3.port['26658/tcp'])

    async with client3.connect(auth_token3) as api:
        address3 = await api.state.account_address()
        start_balance3 = await api.state.balance()
        await api.blob.submit(Blob(b'abc', b'client3'))

    async with client2.connect(auth_token2) as api:
        address2 = await api.state.account_address()
        start_balance2 = await api.state.balance()
        await api.blob.submit(Blob(b'abc', b'client2'))

    async with client1.connect(auth_token1) as api:
        start_balance1 = await api.state.balance()
        await api.blob.submit(Blob(b'abc', b'client1'))
        await asyncio.sleep(5)
        new_balance3 = await api.state.balance_for_address(address3)
        new_balance2 = await api.state.balance_for_address(address2)
        new_balance1 = await api.state.balance()

        assert new_balance3.amount < start_balance3.amount
        assert new_balance2.amount < start_balance2.amount
        assert new_balance1.amount < start_balance1.amount

        await api.state.transfer(address3, 20)
        await api.state.transfer(address2, 20)

        await asyncio.sleep(5)
        assert (await api.state.balance()).amount < new_balance1.amount - 40
        assert (await api.state.balance_for_address(address2)).amount == new_balance2.amount + 20
        assert (await api.state.balance_for_address(address3)).amount == new_balance3.amount + 20

        await api.state.grant_fee(address2, 10000)
        await api.state.revoke_grant_fee(address2)


@pytest.mark.asyncio
async def test_undelegate(node_provider, validator_addresses):
    bridge, auth_token = await node_provider('bridge-0')
    client = Client(port=bridge.port['26658/tcp'])
    validator1 = validator_addresses[0]
    try:
        async with client.connect(auth_token) as api:
            await api.blob.submit(Blob(b'abc', b'client3'))

            try:
                amount_before_delegation = (
                    await api.state.query_delegation(validator1)).delegation_response.balance.amount
            except ValueError:
                amount_before_delegation = 0

            await api.state.delegate(validator1, 10000)
            await api.state.undelegate(validator1, 1000)
            assert (await api.state.query_delegation(
                validator1)).delegation_response.balance.amount == amount_before_delegation + 9000

            query_unbonding1 = await api.state.query_unbonding(validator1)
            undelegate = await api.state.undelegate(validator1, 9000)
            query_unbonding2 = await api.state.query_unbonding(validator1)
            assert len(query_unbonding1.unbond.entries) + 1 == len(query_unbonding2.unbond.entries)

            await api.state.cancel_unbonding_delegation(validator1, 3000, undelegate.height)
            query_unbonding3 = await api.state.query_unbonding(validator1)
            assert query_unbonding2.unbond.entries[-1].balance - 3000 == query_unbonding3.unbond.entries[-1].balance

            query_delegation = await api.state.query_delegation(validator1)
            with pytest.raises(ValueError):
                await api.state.undelegate(validator1, query_delegation.delegation_response.balance.amount + 1000)
            await api.state.undelegate(validator1, query_delegation.delegation_response.balance.amount)

    except ValueError as e:
        if 'too many unbonding delegation entries for' in e.args[0]:
            print(""" The unbound pool is full. To test the functions undelegate, query_unbonding,
                  cancel_unbonding_delegation the network needs to be recreated
                  """
                  )


@pytest.mark.asyncio
async def test_delegating(node_provider, validator_addresses):
    bridge, auth_token = await node_provider('bridge-0')
    client = Client(port=bridge.port['26658/tcp'])
    validator1, validator2, validator3 = validator_addresses
    try:
        async with client.connect(auth_token) as api:
            await api.blob.submit(Blob(b'abc', b'client3'))

            try:
                amount_validator1_before_delegation = (
                    await api.state.query_delegation(validator1)).delegation_response.balance.amount
            except ValueError:
                amount_validator1_before_delegation = 0
            try:
                amount_validator2_before_delegation = (
                    await api.state.query_delegation(validator2)).delegation_response.balance.amount
            except ValueError:
                amount_validator2_before_delegation = 0

            await api.state.delegate(validator1, 10000)
            assert (await api.state.query_delegation(
                validator1)).delegation_response.balance.amount == amount_validator1_before_delegation + 10000

            await api.state.begin_redelegate(validator1, validator2, 10000)
            assert (await api.state.query_delegation(
                validator1)).delegation_response.balance.amount == amount_validator1_before_delegation
            assert (await api.state.query_delegation(
                validator2)).delegation_response.balance.amount == amount_validator2_before_delegation + 10000

            querry = await api.state.query_redelegations(validator1, validator2)
            assert querry.redelegation_responses[0].entries[-1].balance == 10000

            with pytest.raises(ValueError):
                await api.state.begin_redelegate(validator1, validator1, 3000)
            with pytest.raises(ValueError):
                await api.state.query_redelegations(validator1, validator1)
            with pytest.raises(ValueError):
                await api.state.query_redelegations(validator2, validator1)

    except ValueError as e:
        if 'too many redelegation entries for (delegator, src-validator, dst-validator)' in e.args[0]:
            print(""" The redelegation entries is full. To test the functions query_delegation, delegate,
                  begin_redelegate the network needs to be recreated
                  """
                  )



from collections.abc import AsyncIterator
from functools import wraps
from typing import Callable

from celestia.types.header import ExtendedHeader, State
from ._RPC import Wrapper


def handle_header_error(func):
    """ Decorator to handle blob-related errors."""

    @wraps(func)
    async def wrapper(*args, **kwargs):
        try:
            return await func(*args, **kwargs)
        except ConnectionError as e:
            if 'header: not found' in e.args[1].body['message'].lower():
                return None
            raise

    return wrapper


class HeaderClient(Wrapper):
    """ Client for interacting with Celestia's Header API."""

    @handle_header_error
    async def get_by_hash(self, header_hash: str, deserializer: Callable | None = None) -> ExtendedHeader | None:
        """ Returns the header of the given hash from the node's header store.

        Args:
            header_hash (str): The hash of the header to retrieve.
            deserializer (Callable | None): Custom deserializer. Defaults to :meth:`~celestia.types.header.ExtendedHeader.deserializer`.

        Returns:
            ExtendedHeader | None: The retrieved header if found, otherwise None.
        """

        deserializer = deserializer if deserializer is not None else ExtendedHeader.deserializer

        return await self._rpc.call("header.GetByHash", (header_hash,), deserializer)

    async def get_by_height(self, height: int, deserializer: Callable | None = None) -> ExtendedHeader:
        """ Returns the ExtendedHeader at the given height if it is currently available.

        Args:
            height (int): The height of the header.
            deserializer (Callable | None): Custom deserializer. Defaults to :meth:`~celestia.types.header.ExtendedHeader.deserializer`.

        Returns:
            ExtendedHeader: The retrieved header.
        """

        deserializer = deserializer if deserializer is not None else ExtendedHeader.deserializer

        return await self._rpc.call("header.GetByHeight", (int(height),), deserializer)

    async def get_range_by_height(self, range_from: ExtendedHeader, range_to: int,
                                  deserializer: Callable | None = None) -> list[ExtendedHeader]:
        """ Returns the given range (from:to) of ExtendedHeaders from the node's header store
        and verifies that the returned headers are adjacent to each other.

        Args:
            range_from (ExtendedHeader): The starting header.
            range_to (int): The height of the last header in the range.
            deserializer (Callable | None): Custom deserializer. Defaults to None.

        Returns:
            list[ExtendedHeader]: A list of retrieved headers.
        """

        def deserializer_(result):
            if result is not None:
                return [ExtendedHeader(**kwargs) for kwargs in result]

        deserializer = deserializer if deserializer is not None else deserializer_

        return await self._rpc.call("header.GetRangeByHeight", (range_from, int(range_to)), deserializer)

    async def local_head(self, deserializer: Callable | None = None) -> ExtendedHeader:
        """ Returns the ExtendedHeader of the chain head.

        Args:
            deserializer (Callable | None): Custom deserializer. Defaults to :meth:`~celestia.types.header.ExtendedHeader.deserializer`.

        Returns:
            ExtendedHeader: The latest known header of the local node.
        """

        deserializer = deserializer if deserializer is not None else ExtendedHeader.deserializer

        return await self._rpc.call("header.LocalHead", (), deserializer)

    async def network_head(self, deserializer: Callable | None = None) -> ExtendedHeader:
        """ Provides the Syncer's view of the current network head.

        Args:
            deserializer (Callable | None): Custom deserializer. Defaults to :meth:`~celestia.types.header.ExtendedHeader.deserializer`.

        Returns:
            ExtendedHeader: The latest known header of the network.
        """

        deserializer = deserializer if deserializer is not None else ExtendedHeader.deserializer

        return await self._rpc.call("header.NetworkHead", (), deserializer)

    async def subscribe(self, deserializer: Callable | None = None) -> AsyncIterator[ExtendedHeader | None]:
        """ Subscribes to recent ExtendedHeaders from the network.

        Args:
            deserializer (Callable | None): Custom deserializer. Defaults to :meth:`~celestia.types.header.ExtendedHeader.deserializer`.

        Yields:
            ExtendedHeader | None: The latest headers as they become available.
        """

        deserializer = deserializer if deserializer is not None else ExtendedHeader.deserializer

        async for subs_header_result in self._rpc.iter("header.Subscribe", (), deserializer):
            if subs_header_result is not None:
                yield subs_header_result

    async def sync_state(self, deserializer: Callable | None = None) -> State:
        """ Returns the current state of the header Syncer.

        Args:
            deserializer (Callable | None): Custom deserializer. Defaults to :meth:`~celestia.types.header.State.deserializer`.

        Returns:
           State: The current synchronization state.
        """

        deserializer = deserializer if deserializer is not None else State.deserializer

        return await self._rpc.call("header.SyncState", (), deserializer)

    async def sync_wait(self) -> None:
        """ Blocks until the header Syncer is synced to network head.

        Returns:
            None
        """
        return await self._rpc.call("header.SyncWait")

    async def wait_for_height(self, height: int, deserializer: Callable | None = None) -> ExtendedHeader:
        """ Blocks until the header at the given height has been processed
        by the store or context deadline is exceeded.

        Args:
            height (int): The height of the header to wait for.
            deserializer (Callable | None): Custom deserializer. Defaults to :meth:`~celestia.types.header.ExtendedHeader.deserializer`.

        Returns:
            ExtendedHeader: The retrieved header once available.
        """

        deserializer = deserializer if deserializer is not None else ExtendedHeader.deserializer

        return await self._rpc.call("header.WaitForHeight", (height,), deserializer)



import typing as t
from base64 import b64decode, b64encode
from dataclasses import dataclass

from celestia._celestia import types as ext  # noqa


class Base64(bytes):
    """ Represents a byte string that supports Base64 encoding and decoding.

    This class ensures that the stored data is always in bytes and provides
    Base64 encoding/decoding when converting to and from strings.
    """

    def __new__(cls, value: str | bytes):
        if isinstance(value, str):
            value = b64decode(value)
        if value is None:
            return None
        return super().__new__(cls, value)

    def __str__(self) -> str:
        return b64encode(self).decode('ascii')

    @classmethod
    def ensure_type(cls, value):
        """ Ensures the value is an instance of Base64.

        Args:
            value (str | bytes | Base64): The value to convert.

        Returns:
            Base64: A valid Base64 instance.
        """
        if isinstance(value, cls):
            return value
        return cls(value)


class Namespace(Base64):
    """ Represents a Celestia namespace.

    A namespace is a unique identifier for blobs stored on the Celestia network.
    It is used to segregate data based on different applications or use cases.
    """

    def __new__(cls, value: str | bytes):
        value = super().__new__(cls, value)
        value = ext.normalize_namespace(value)
        return super().__new__(cls, value)


class Commitment(Base64):
    """ Represents a Celestia blob commitment.

    A commitment is a cryptographic proof that ensures data integrity and allows
    verification of whether a blob has been correctly included in a block.
    """


@dataclass
class Blob:
    """ Represents a Celestia blob.

    A blob is a chunk of data stored on Celestia. Each blob is associated with
    a namespace and a cryptographic commitment to ensure data integrity.

    Attributes:
        namespace (Namespace): The namespace under which the blob is stored.
        data (Base64): The actual blob data.
        commitment (Commitment): The cryptographic commitment for the blob.
        share_version (int): The version of the share encoding used.
        index (int | None): The index of the blob in the block (optional).
    """
    namespace: Namespace
    data: Base64
    commitment: Commitment
    share_version: int
    index: int | None = None

    def __init__(self, namespace: Namespace | str | bytes, data: Base64 | str | bytes,
                 commitment: Commitment | str | bytes | None = None, share_version: int | None = 0,
                 index: int | None = None):
        self.namespace = Namespace.ensure_type(namespace)
        self.data = Base64.ensure_type(data)
        if commitment is not None:
            self.commitment = Commitment.ensure_type(commitment)
            self.share_version = share_version or 0
        else:
            kwargs = ext.normalize_blob(self.namespace, self.data)
            self.commitment = Commitment(kwargs['commitment'])
            self.share_version = kwargs['share_version']
        self.index = index

    @staticmethod
    def deserializer(result: dict) -> 'Blob':
        """ Deserializes a dictionary into a Blob object.

        Args:
            result: The dictionary representation of a Blob.

        Returns:
            A deserialized Blob object.
        """
        if result is not None:
            return Blob(**result)


class TxConfig(t.TypedDict):
    """ Represents a transaction configuration for submitting transactions to Celestia.

    Attributes:
        signer_address (str | None): The address of the transaction signer.
        is_gas_price_set (bool | None): Whether a custom gas price is set.
        key_name (str | None): The name of the key used for signing.
        gas_price (float | None): The gas price for the transaction.
        gas (int | None): The amount of gas to use.
        fee_granter_address (str | None): Address of the fee granter (if applicable).
    """
    signer_address: str | None
    is_gas_price_set: bool | None
    key_name: str | None
    gas_price: float | None
    gas: int | None
    fee_granter_address: str | None



import pytest

from celestia.node_api import Client


@pytest.mark.asyncio
async def test_testnet(node_provider):
    bridge, auth_token = await node_provider('bridge-0')
    client = Client(port=bridge.port['26658/tcp'])
    async with client.connect(auth_token) as api:
        balance = await api.state.balance()
        assert balance.amount



from dataclasses import dataclass

from celestia._celestia import types as ext  # noqa

from celestia.types.common_types import Blob, Base64, Namespace, Commitment


@dataclass
class SubmitBlobResult:
    """ Represents the result of submitting a blob to the Celestia network.

    Attributes:
        height (int): The block height at which the blob was submitted.
        commitments (tuple[Commitment, ...]): Commitments associated with the submitted blob.
    """
    height: int
    commitments: tuple[Commitment, ...]


@dataclass
class SubscriptionBlobResult:
    """ Represents the result of a subscription to blobs in the Celestia network.

    Attributes:
        height (int): The block height of the retrieved blobs.
        blobs (tuple[Blob, ...]): The list of blobs retrieved from the subscription.
    """
    height: int
    blobs: tuple[Blob, ...]


@dataclass
class Proof:
    """ Represents a Merkle proof used for verifying data inclusion in Celestia.

    Attributes:
        end (int): The end index of the proof range.
        nodes (tuple[Base64, ...]): The nodes forming the Merkle proof.
        start (int | None): The start index of the proof range (optional).
        is_max_namespace_ignored (bool | None): Flag indicating if max namespace check is ignored (optional).
    """
    end: int
    nodes: tuple[Base64, ...]
    start: int | None
    is_max_namespace_ignored: bool | None

    def __init__(self, nodes, end, is_max_namespace_ignored=None, start=None):
        self.start = start
        self.nodes = tuple(node for node in nodes)
        self.end = end
        self.is_max_namespace_ignored = is_max_namespace_ignored

    @staticmethod
    def deserializer(result) -> list['Proof']:
        """ Deserializes a list of proof entries from a given result.

        Args:
            result (list[dict]): The serialized response data.

        Returns:
            list[Proof]: A list of deserialized Proof objects.
        """
        if result is not None:
            return [Proof(**kwargs) for kwargs in result]


@dataclass
class RowProofEntry:
    """ Represents an entry in a row proof, used for verifying inclusion in a specific row of a Merkle tree.

    Attributes:
        index (int | None): The index of the leaf in the row.
        total (int): The total number of leaves in the row.
        leaf_hash (Base64): The hash of the leaf.
        aunts (tuple[Base64, ...]): The sibling hashes used in the proof.
    """
    index: int | None
    total: int
    leaf_hash: Base64
    aunts: tuple[Base64, ...]

    def __init__(self, leaf_hash, aunts, total, index=None):
        self.leaf_hash = leaf_hash
        self.aunts = tuple(aunt for aunt in aunts)
        self.total = total
        self.index = index


@dataclass
class RowProof:
    """ Represents a proof for a row in a Merkle tree.

    Attributes:
        start_row (int | None): The starting row index of the proof.
        end_row (int | None): The ending row index of the proof.
        row_roots (tuple[Base64, ...]): The root hashes of the rows.
        proofs (tuple[RowProofEntry, ...]): The proof entries for the row.
    """
    start_row: int | None
    end_row: int | None
    row_roots: tuple[Base64, ...]
    proofs: tuple[RowProofEntry, ...]

    def __init__(self, row_roots, proofs, end_row=None, start_row=None):
        self.row_roots = tuple(row_root for row_root in row_roots)
        self.proofs = tuple(RowProofEntry(**proof) for proof in proofs)
        self.end_row = end_row
        self.start_row = start_row


@dataclass
class CommitmentProof:
    """ Represents a proof of commitment in Celestia, verifying that a namespace is correctly included.

    Attributes:
        namespace_id (Namespace): The namespace identifier.
        namespace_version (int): The version of the namespace.
        row_proof (RowProof): The proof for the rows containing the namespace.
        subtree_root_proofs (tuple[Proof, ...]): Proofs for verifying subtree roots.
        subtree_roots (tuple[Base64, ...]): The roots of the subtrees.
    """
    namespace_id: Namespace
    namespace_version: int
    row_proof: RowProof
    subtree_root_proofs: tuple[Proof, ...]
    subtree_roots: tuple[Base64, ...]

    def __init__(self, namespace_id, namespace_version, row_proof, subtree_root_proofs, subtree_roots):
        self.namespace_id = Namespace.ensure_type(namespace_id)
        self.namespace_version = int(namespace_version)
        self.row_proof = RowProof(**row_proof)
        self.subtree_root_proofs = tuple(Proof(**subtree_root_proof) for subtree_root_proof in subtree_root_proofs)
        self.subtree_roots = tuple(subtree_root for subtree_root in subtree_roots)

    @staticmethod
    def deserializer(result: dict) -> 'CommitmentProof':
        """ Deserializes a commitment proof from a given result.

        Args:
            result (dict): The dictionary representation of a CommitmentProof.

        Returns:
            A deserialized CommitmentProof object.
        """
        if result is not None:
            return CommitmentProof(**result)



import asyncio
import json
import typing as t
import uuid
from asyncio import Future
from collections import deque
from collections.abc import AsyncGenerator
from contextlib import asynccontextmanager, AbstractAsyncContextManager
from dataclasses import is_dataclass, asdict

from ajsonrpc.core import JSONRPC20Response, JSONRPC20Request

from celestia.types.common_types import Base64

RPC_VALUE_ERRORS = [
    'unmarshaling params',
    'equal to 0',
    'given height is from the future',
    'invalid range',
    'height must be bigger than zero',
    'dial to self attempted',
    'gater disallows connection to peer',
    'notfound desc = delegation with delegator',
    'unknown desc = failed to execute message; message index: 0: invalid shares amount:',
    'cannot redelegate to the same validator',
    'too many unbonding delegation entries for (delegator, validator) tuple',
    'redelegation not found for delegator address',
    'too many redelegation entries for (delegator, src-validator, dst-validator)',
    'datastore: key not found'
]


class JSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if is_dataclass(obj):
            return asdict(obj)
        if isinstance(obj, Base64):
            return str(obj)
        return super().default(obj)


class Wrapper:
    def __init__(self, rpc: 'RPC'):
        self._rpc = rpc


class Transport(t.Protocol):

    async def send(self, message: str) -> None:
        """ Send a message to the connection. """


class RPC:
    """ RPC executor """

    def __init__(self, response_timeout: int):
        self._response_timeout = response_timeout
        self._transport = None  # type: Transport | None
        self._pending = dict()  # type: dict[str, Future]
        self._subscriptions = dict()  # type: dict[str,deque[t.Any]]

    @property
    def transport(self) -> Transport:
        if self._transport:
            return self._transport
        raise ConnectionError("Transport is not connected.")

    def connect(self, transport: Transport) -> AbstractAsyncContextManager[t.Callable[[str], None]]:

        def receiver(message: str):
            message = json.loads(message)
            if 'method' in message:
                subscription_id, item = message['params']
                subscription = self._subscriptions.get(subscription_id, None)
                if subscription is not None:
                    subscription.append(item)
            else:
                response = JSONRPC20Response(result=False)
                response.body = message
                if future := self._pending.get(response.id):
                    if response.error is not None:
                        error_body = getattr(response.error, "body", None)
                        error_message = error_body.get('message', None).lower() if error_body else None
                        if any(keyword in error_message for keyword in RPC_VALUE_ERRORS):
                            future.set_exception(ValueError(error_message))
                        elif error_message is None or error_body is None:
                            future.set_exception(ConnectionError("RPC faled; undefined error"))
                        else:
                            future.set_exception(ConnectionError(f"RPC failed; {error_message}", response.error))
                    else:
                        future.set_result(response.result)
                else:
                    raise RuntimeError("Received message with unexpected ID.")

        @asynccontextmanager
        async def connect_context():
            try:
                self._transport = transport
                yield receiver
            finally:
                self._transport = None
                for id in tuple(self._pending.keys()):
                    future = self._pending.pop(id)
                    if not future.done():
                        self._pending[id].set_exception(ConnectionError("RPC closed"))
                self._pending.clear()
                # ToDo: cleanup subscription

        return connect_context()

    async def call(self, method: str, params: tuple[t.Any, ...] = None,
                   deserializer: t.Callable[[t.Any], t.Any] = None) -> t.Any | None:
        params = params or ()
        deserializer = deserializer or (lambda a: a)
        id = str(uuid.uuid4())
        request = JSONRPC20Request(method, params, id)
        await self._transport.send(json.dumps(request.body, cls=JSONEncoder))
        future = self._pending[id] = Future()
        future.add_done_callback(lambda _: self._pending.pop(id, None))
        async with asyncio.timeout(self._response_timeout):
            result = await future
            return deserializer(result)

    async def iter(self, method: str, params: tuple[t.Any, ...] = None,
                   deserializer: t.Callable[[t.Any], t.Any] = None
                   ) -> AsyncGenerator[t.Any]:
        deserializer = deserializer or (lambda a: a)
        subscription_id = await self.call(method, params)
        try:
            subscription = self._subscriptions[subscription_id] = deque()
            while True:
                if len(subscription):
                    yield deserializer(subscription.popleft())
                else:
                    await asyncio.sleep(0.1)
        finally:
            del self._subscriptions[subscription_id]



import typing as t
from dataclasses import dataclass

from celestia._celestia import types as ext  # noqa


@dataclass
class Balance:
    """ Represents the balance of a particular denomination.

    Attributes:
        amount (int): The amount of the balance.
        denom (str): The denomination of the balance.
    """
    amount: int
    denom: str

    def __init__(self, amount: int, denom: str):
        self.amount = int(amount)
        self.denom = denom

    @staticmethod
    def deserializer(result: dict) -> 'Balance':
        """ Deserialize a result dictionary into a Balance object.

        Args:
            result (dict): The dictionary representation of a Balance.

        Returns:
            Balance: The deserialized Balance object.
        """
        if result is not None:
            return Balance(**result)


@dataclass
class TXResponse:
    """ Represents the response for a transaction.

    Attributes:
        height (int): The block height of the transaction.
        txhash (str): The transaction hash.
        logs (tuple[t.Any] | None): Logs associated with the transaction, if any.
        events (tuple[t.Any, ...] | None): Events triggered by the transaction, if any.
    """
    height: int
    txhash: str
    logs: tuple[t.Any] | None = None
    events: tuple[t.Any, ...] | None = None

    def __init__(self, height, txhash, logs, events):
        self.height = int(height)
        self.txhash = txhash
        self.logs = tuple(log for log in logs) if logs else None
        self.events = tuple(event for event in events) if events else None

    @staticmethod
    def deserializer(result: dict) -> 'TXResponse':
        """ Deserialize a result dictionary into a TXResponse object.

        Args:
            result (dict): The dictionary representation of a TXResponse.

        Returns:
            TXResponse: The deserialized TXResponse object.
        """
        if result is not None:
            return TXResponse(**result)


@dataclass
class Delegation:
    """ Represents a delegation of tokens to a validator.

    Attributes:
        delegator_address (str): The address of the delegator.
        validator_address (str): The address of the validator.
        shares (float): The amount of shares in the delegation.
    """
    delegator_address: str
    validator_address: str
    shares: float

    def __init__(self, delegator_address, validator_address, shares):
        self.delegator_address = delegator_address
        self.validator_address = validator_address
        self.shares = float(shares)


@dataclass
class DelegationResponse:
    """ Represents the response for a delegation query.

    Attributes:
       delegation (Delegation): The delegation details.
       balance (Balance): The balance associated with the delegation.
    """
    delegation: Delegation
    balance: Balance

    def __init__(self, delegation, balance):
        self.delegation = Delegation(**delegation)
        self.balance = Balance(**balance)


@dataclass
class QueryDelegationResponse:
    """ Represents the response for a delegation query.

    Attributes:
        delegation_response (DelegationResponse): The delegation response details.
    """
    delegation_response: DelegationResponse

    def __init__(self, delegation_response):
        self.delegation_response = DelegationResponse(**delegation_response)

    @staticmethod
    def deserializer(result: dict) -> 'QueryDelegationResponse':
        """ Deserialize a result dictionary into a QueryDelegationResponse object.

        Args:
            result (dict): The dictionary representation of a QueryDelegationResponse.

        Returns:
            QueryDelegationResponse: The deserialized QueryDelegationResponse object.
        """
        if result is not None:
            return QueryDelegationResponse(**result)


@dataclass
class RedelegationEntry:
    """ Represents a redelegation entry.

    Attributes:
       creation_height (int): The block height when the redelegation was created.
       completion_time (str): The completion time of the redelegation.
       initial_balance (int): The initial balance of the redelegation.
       shares_dst (float): The amount of shares transferred to the destination validator.
    """
    creation_height: int
    completion_time: str
    initial_balance: int
    shares_dst: float

    def __init__(self, creation_height, completion_time, initial_balance, shares_dst):
        self.creation_height = creation_height
        self.completion_time = completion_time
        self.initial_balance = int(initial_balance)
        self.shares_dst = float(shares_dst)


@dataclass
class Redelegation:
    """ Represents a redelegation of tokens from one validator to another.

    Attributes:
        delegator_address (str): The address of the delegator.
        validator_src_address (str): The address of the source validator.
        validator_dst_address (str): The address of the destination validator.
        entries (tuple[RedelegationEntry, ...]): A list of redelegation entries.
    """
    delegator_address: str
    validator_src_address: str
    validator_dst_address: str
    entries: tuple[RedelegationEntry, ...]

    def __init__(self, delegator_address, validator_src_address, validator_dst_address, entries):
        self.delegator_address = delegator_address
        self.validator_src_address = validator_src_address
        self.validator_dst_address = validator_dst_address
        self.entries = tuple(RedelegationEntry(**entry) for entry in entries)


@dataclass
class RedelegationResponseEntry:
    """ Represents a redelegation response entry.

    Attributes:
        redelegation_entry (RedelegationEntry): The redelegation entry details.
        balance (int): The balance of the redelegation entry.
    """
    redelegation_entry: RedelegationEntry
    balance: int

    def __init__(self, redelegation_entry, balance):
        self.redelegation_entry = RedelegationEntry(**redelegation_entry)
        self.balance = int(balance)


@dataclass
class RedelegationResponse:
    """ Represents the response for a redelegation query.

    Attributes:
        redelegation (Redelegation): The redelegation details.
        entries (tuple[RedelegationResponseEntry, ...]): A list of redelegation response entries.
    """
    redelegation: Redelegation
    entries: tuple[RedelegationResponseEntry, ...]

    def __init__(self, redelegation, entries):
        self.redelegation = Redelegation(**redelegation)
        self.entries = tuple(RedelegationResponseEntry(**entry) for entry in entries)


@dataclass
class Pagination:
    """ Represents pagination information.

    Attributes:
        next_key (str | None): The key for the next page of results.
        total (int | None): The total number of results.
    """
    next_key: str = None
    total: int = None


@dataclass
class QueryRedelegationResponse:
    """ Represents the response for a query to retrieve redelegations.

    Attributes:
        redelegation_responses (tuple[RedelegationResponse, ...]): A list of redelegation responses.
        pagination (Pagination): Pagination information for the query results.
    """
    redelegation_responses: tuple[RedelegationResponse, ...]
    pagination: Pagination

    def __init__(self, redelegation_responses, pagination=None):
        self.redelegation_responses = tuple(
            RedelegationResponse(**redelegation_response) for redelegation_response in redelegation_responses)
        self.pagination = Pagination(**pagination) if pagination else None

    @staticmethod
    def deserializer(result: dict) -> 'QueryRedelegationResponse':
        """ Deserialize a result dictionary into a QueryRedelegationResponse object.

        Args:
            result (dict): The dictionary representation of a QueryRedelegationResponse.

        Returns:
            QueryRedelegationResponse: The deserialized QueryRedelegationResponse object.
        """
        if result is not None:
            return QueryRedelegationResponse(**result)


@dataclass
class UnbondEntry:
    """ Represents an unbonding entry for a validator.

    Attributes:
        creation_height (int): The block height when the unbonding was created.
        completion_time (str): The completion time of the unbonding.
        initial_balance (int): The initial balance of the unbonding.
        balance (int): The current balance after unbonding.
    """
    creation_height: int
    completion_time: str
    initial_balance: int
    balance: int

    def __init__(self, creation_height, completion_time, initial_balance, balance):
        self.creation_height = creation_height
        self.completion_time = completion_time
        self.initial_balance = int(initial_balance)
        self.balance = int(balance)


@dataclass
class Unbond:
    """ Represents an unbonding of tokens from a validator.

    Attributes:
        delegator_address (str): The address of the delegator.
        validator_address (str): The address of the validator.
        entries (tuple[UnbondEntry, ...]): A list of unbonding entries.
    """
    delegator_address: str
    validator_address: str
    entries: tuple[UnbondEntry, ...]

    def __init__(self, delegator_address, validator_address, entries):
        self.delegator_address = delegator_address
        self.validator_address = validator_address
        self.entries = tuple(UnbondEntry(**entry) for entry in entries)


@dataclass
class QueryUnbondingDelegationResponse:
    """ Represents the response for a query to retrieve unbonding delegations.

    Attributes:
        unbond (Unbond): The unbonding details.
    """
    unbond: Unbond

    def __init__(self, unbond):
        self.unbond = Unbond(**unbond)

    @staticmethod
    def deserializer(result: dict) -> 'QueryUnbondingDelegationResponse':
        """ Deserialize a result dictionary into a QueryUnbondingDelegationResponse object.

        Args:
            result (dict): The dictionary representation of a QueryUnbondingDelegationResponse.

        Returns:
            QueryUnbondingDelegationResponse: The deserialized QueryUnbondingDelegationResponse object.
        """
        if result is not None:
            return QueryUnbondingDelegationResponse(**result)



import asyncio

import pytest

from celestia.node_api import Client


@pytest.mark.asyncio
async def test_header(node_provider):
    bridge, auth_token = await node_provider('bridge-0')
    client = Client(port=bridge.port['26658/tcp'])

    async with client.connect(auth_token) as api:
        local_head = await api.header.local_head()
        network_head = await api.header.network_head()
        local_height = local_head.header.height
        local_hash = local_head.commit.block_id.hash
        assert local_height <= network_head.header.height

        head = await api.header.get_by_hash('4D3818BC5D3BE8E529C953C8654BD4243A2CD28BD1599DBF0ED4DD44C24F6D33')
        assert head is None
        head = await api.header.get_by_hash(local_hash)
        assert local_head == head

        head = await api.header.get_by_height(local_height)
        assert local_head == head

        heads = await api.header.get_range_by_height(await api.header.get_by_height(int(local_height) - 5),
                                                     local_height)
        assert len(heads) == 4

        state1 = await api.header.sync_state()

        await api.header.wait_for_height(state1.height + 1)

        state2 = await api.header.sync_state()
        assert state1.height <= state2.height


@pytest.mark.asyncio
async def test_header_exceptions(node_provider):
    bridge, auth_token = await node_provider('bridge-0')
    client = Client(port=bridge.port['26658/tcp'])
    async with client.connect(auth_token) as api:
        local_head = await api.header.local_head()
        local_height = local_head.header.height

        with pytest.raises(ValueError):
            await api.header.get_by_height(18446744073709551615)
        with pytest.raises(ValueError):
            await api.header.get_by_height(0)
        with pytest.raises(ValueError):
            await api.header.get_by_height(-1)

        with pytest.raises(ValueError):
            await api.header.get_range_by_height(await api.header.get_by_height(local_height), int(local_height) - 5)

        with pytest.raises(ValueError):
            await api.header.wait_for_height(0)
        with pytest.raises(ValueError):
            await api.header.wait_for_height(-1)


@pytest.mark.asyncio
async def test_header_subscribe(node_provider):
    bridge, auth_token = await node_provider('bridge-0')
    client = Client(port=bridge.port['26658/tcp'])
    result = []

    async with client.connect(auth_token) as api:
        async with asyncio.timeout(30):
            async for header in api.header.subscribe():
                result.append(header)
                if len(result) == 3:
                    break
    assert len(result) == 3



from typing import Callable

from celestia.types.p2p import BandwidthStats, Connectedness, AddrInfo, Reachability, ResourceManagerStat
from ._RPC import Wrapper


class P2PClient(Wrapper):
    """ Client for interacting with Celestia's P2P API."""

    async def bandwidth_for_peer(self, peer_id: str, deserializer: Callable | None = None) -> BandwidthStats:
        """ Returns a Stats struct with bandwidth metrics associated with the given peer.ID.
        The metrics returned include all traffic sent / received for the peer, regardless of protocol.

        Args:
            peer_id (str): The ID of the peer.
            deserializer (Callable | None): Custom deserializer. Defaults to :meth:`~celestia.types.p2p.BandwidthStats.deserializer`.

        Returns:
            BandwidthStats: Bandwidth statistics for the given peer.
        """

        deserializer = deserializer if deserializer is not None else BandwidthStats.deserializer

        return await self._rpc.call("p2p.BandwidthForPeer", (peer_id,), deserializer)

    async def bandwidth_for_protocol(self, protocol_id: str, deserializer: Callable | None = None) -> BandwidthStats:
        """ Returns a Stats struct with bandwidth metrics associated with the given protocol.ID.

        Args:
            protocol_id (str): The protocol ID.
            deserializer (Callable | None): Custom deserializer. Defaults to :meth:`~celestia.types.p2p.BandwidthStats.deserializer`.

        Returns:
            BandwidthStats: Bandwidth statistics for the given protocol.
        """

        deserializer = deserializer if deserializer is not None else BandwidthStats.deserializer

        return await self._rpc.call("p2p.BandwidthForProtocol", (protocol_id,), deserializer)

    async def bandwidth_stats(self, deserializer: Callable | None = None) -> BandwidthStats:
        """ Returns a Stats struct with bandwidth metrics for all data sent/received by the local peer,
        regardless of protocol or remote peer IDs.

        Args:
            deserializer (Callable | None): Custom deserializer. Defaults to :meth:`~celestia.types.p2p.BandwidthStats.deserializer`.

        Returns:
            BandwidthStats: Overall bandwidth statistics.
        """

        deserializer = deserializer if deserializer is not None else BandwidthStats.deserializer

        return await self._rpc.call("p2p.BandwidthStats", (), deserializer)

    async def block_peer(self, peer_id: str) -> None:
        """ Adds a peer to the set of blocked peers and closes any existing connection to that peer.

        Args:
            peer_id (str): The ID of the peer to block.
        """
        await self._rpc.call("p2p.BlockPeer", (peer_id,))

    async def close_peer(self, peer_id: str) -> None:
        """ Closes the connection to a given peer.

        Args:
            peer_id (str): The ID of the peer to disconnect from.
        """
        await self._rpc.call("p2p.ClosePeer", (peer_id,))

    async def connect(self, address: AddrInfo) -> None:
        """ Ensures there is a connection between this host and the peer with given peer.

        Args:
            address (AddrInfo): Address information of the peer.
        """
        await self._rpc.call("p2p.Connect", (address,))

    async def connectedness(self, peer_id: str) -> Connectedness:
        """ Returns a state signaling connection capabilities.

        Args:
            peer_id (str): The ID of the peer.

        Returns:
            Connectedness: Connection status with the peer.
        """
        return await self._rpc.call("p2p.Connectedness", (peer_id,))

    async def info(self, deserializer: Callable | None = None) -> AddrInfo:
        """ Returns address information about the host.

        Args:
            deserializer (Callable | None): Custom deserializer. Defaults to :meth:`~celestia.types.p2p.AddrInfo.deserializer`.

        Returns:
            AddrInfo: Address information of the local peer.
        """

        deserializer = deserializer if deserializer is not None else AddrInfo.deserializer

        return await self._rpc.call("p2p.Info", (), deserializer)

    async def is_protected(self, peer_id: str, tag: str) -> bool:
        """ Returns whether the given peer is protected.

        Args:
            peer_id (str): The ID of the peer.
            tag (str): Protection tag.

        Returns:
            bool: True if the peer is protected, False otherwise.
        """
        return await self._rpc.call("p2p.IsProtected", (peer_id, tag))

    async def list_blocked_peers(self) -> list[str]:
        """ Returns a list of blocked peers.

        Returns:
            list[str]: A list of blocked peer IDs.
        """
        return await self._rpc.call("p2p.ListBlockedPeers")

    async def nat_status(self) -> Reachability:
        """ Returns the current NAT status.

        Returns:
            Reachability: NAT reachability status.
        """
        return await self._rpc.call("p2p.NATStatus")

    async def peer_info(self, peer_id: str, deserializer: Callable | None = None) -> AddrInfo:
        """ Returns a small slice of information Peerstore has on the given peer.

        Args:
            peer_id (str): The ID of the peer.
            deserializer (Callable | None): Custom deserializer. Defaults to :meth:`~celestia.types.p2p.AddrInfo.deserializer`.

        Returns:
            AddrInfo: Address information of the peer.
        """

        deserializer = deserializer if deserializer is not None else AddrInfo.deserializer

        return await self._rpc.call("p2p.PeerInfo", (peer_id,), deserializer)

    async def peers(self) -> list[str]:
        """ Returns connected peers.

        Returns:
            list[str]: List of connected peer IDs.
        """
        return await self._rpc.call("p2p.Peers")

    async def protect(self, peer_id: str, tag: str) -> None:
        """ Adds a peer to the list of peers who have a bidirectional peering agreement
        that they are protected from being trimmed, dropped or negatively scored.

        Args:
            peer_id (str): The ID of the peer.
            tag (str): Protection tag.
        """
        await self._rpc.call("p2p.Protect", (peer_id, tag))

    async def pub_sub_peers(self, topic: str) -> list[str]:
        """ Returns the peer IDs of the peers joined on the given topic.

        Args:
            topic (str): The PubSub topic to query.

        Returns:
            list[str]: A list of peer IDs that are joined on the specified topic.
        """
        return await self._rpc.call("p2p.PubSubPeers", (topic,))

    async def pub_sub_topics(self) -> list[str] | None:
        """ Reports current PubSubTopics the node participates in.

        Returns:
            list[str] | None: A list of topic names if available, otherwise None.
        """
        return await self._rpc.call("p2p.PubSubTopics")

    async def resource_state(self, deserializer: Callable | None = None) -> ResourceManagerStat:
        """ Returns the state of the resource manager.

        Args:
            deserializer (Callable | None): Custom deserializer. Defaults to :meth:`~celestia.types.p2p.ResourceManagerStat.deserializer`.

        Returns:
            ResourceManagerStat: Resource manager state.
        """

        deserializer = deserializer if deserializer is not None else ResourceManagerStat.deserializer

        return await self._rpc.call("p2p.ResourceState", (), deserializer)

    async def unblock_peer(self, peer_id: str) -> None:
        """ Removes a peer from the set of blocked peers.

        Args:
            peer_id (str): The ID of the peer to unblock.
        """
        await self._rpc.call("p2p.UnblockPeer", (peer_id,))

    async def unprotect(self, peer_id: str, tag: str) -> bool:
        """ Removes a peer from the list of peers who have a bidirectional peering agreement that
        they are protected from being trimmed, dropped or negatively scored, returning a bool
        representing whether the given peer is protected or not.

        Args:
            peer_id (str): The ID of the peer.
            tag (str): Protection tag.

        Returns:
            bool: True if the peer remains protected, False otherwise.
        """
        return await self._rpc.call("p2p.Unprotect", (peer_id, tag))



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'pyCelestia'
copyright = '2025, Alesh, Tnmoxa'
author = 'Alesh, Tnmoxa'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = ['sphinx.ext.autodoc',
              'sphinx.ext.napoleon',
              'sphinx.ext.viewcode',
              'sphinx.ext.intersphinx',]

napoleon_google_docstring = True
autodoc_typehints = 'none'
viewcode_follow_imported_members=True

templates_path = ['_templates']
exclude_patterns = []

intersphinx_mapping = {
    'python': ('https://docs.python.org/3', None),
}

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_static_path = ['_static']

html_theme = 'sphinx_rtd_theme'
html_theme_options = {
    "navigation_depth": 2
}



from typing import Callable, Unpack

from celestia._celestia import types  # noqa

from celestia.types.common_types import TxConfig, Blob
from celestia.types.state import Balance, TXResponse, QueryUnbondingDelegationResponse, \
    QueryDelegationResponse, QueryRedelegationResponse
from ._RPC import Wrapper


class StateClient(Wrapper):
    """ Client for interacting with Celestia's State API."""

    async def account_address(self) -> str:
        """ Retrieves the address of the node's account/signer

        Returns:
            str: The address of the node's account.
        """
        return await self._rpc.call("state.AccountAddress")

    async def balance(self, deserializer: Callable | None = None) -> Balance:
        """ Retrieves the Celestia coin balance for the node's account/signer
        and verifies it against the corresponding block's AppHash.

        Args:
            deserializer (Callable | None): Custom deserializer. Defaults to :meth:`~celestia.types.state.Balance.deserializer`.

        Returns:
            Balance: The balance of the node's account.
        """

        deserializer = deserializer if deserializer is not None else Balance.deserializer

        return await self._rpc.call("state.Balance", (), deserializer)

    async def balance_for_address(self, address: str, deserializer: Callable | None = None) -> Balance:
        """ Retrieves the Celestia coin balance for the given address and verifies the returned balance
        against the corresponding block's AppHash. NOTE: the balance returned is the balance reported
        by the block right before the node's current head (head-1). This is due to the fact that for
        block N, the block's `AppHash` is the result of applying the previous block's transaction list.

        Args:
            address (str): The address to query balance for.
            deserializer (Callable | None): Custom deserializer. Defaults to :meth:`~celestia.types.state.Balance.deserializer`.

        Returns:
            Balance: The balance of the given address.
        """

        deserializer = deserializer if deserializer is not None else Balance.deserializer

        return await self._rpc.call("state.BalanceForAddress", (address,), deserializer)

    async def begin_redelegate(self, src_val_addr: str, dst_val_addr: str, amount: int,
                               deserializer: Callable | None = None, **config: Unpack[TxConfig]) -> TXResponse:
        """ Sends a user's delegated tokens to a new validator for redelegation.

        Args:
            src_val_addr (str): Source validator address.
            dst_val_addr (str): Destination validator address.
            amount (int): Amount to redelegate.
            deserializer (Callable | None): Custom deserializer. Defaults to :meth:`~celestia.types.state.TXResponse.deserializer`.
            **config(TxConfig): Additional transaction configuration.

        Returns:
            TXResponse: Transaction response.
        """

        deserializer = deserializer if deserializer is not None else TXResponse.deserializer

        return await self._rpc.call("state.BeginRedelegate", (src_val_addr, dst_val_addr, str(amount), config),
                                    deserializer)

    async def cancel_unbonding_delegation(self, val_addr: str, amount: int, height: int,
                                          deserializer: Callable | None = None,
                                          **config: Unpack[TxConfig]) -> TXResponse:
        """ Cancels a user's pending undelegation from a validator.

        Args:
            val_addr (str): Validator address.
            amount (int): Amount to cancel unbonding.
            height (int): Block height at which unbonding was initiated.
            deserializer (Callable | None): Custom deserializer. Defaults to :meth:`~celestia.types.state.TXResponse.deserializer`.
            **config(TxConfig): Additional transaction configuration.

        Returns:
            TXResponse: Transaction response.
        """

        deserializer = deserializer if deserializer is not None else TXResponse.deserializer

        return await self._rpc.call("state.CancelUnbondingDelegation", (val_addr, str(amount), str(height), config),
                                    deserializer)

    async def delegate(self, del_addr: str, amount: int, deserializer: Callable | None = None,
                       **config: Unpack[TxConfig]) -> TXResponse:
        """ Sends a user's liquid tokens to a validator for delegation.

        Args:
            del_addr (str): Delegator address.
            amount (int): Amount to delegate.
            deserializer (Callable | None): Custom deserializer. Defaults to :meth:`~celestia.types.state.TXResponse.deserializer`.
            **config(TxConfig): Additional transaction configuration.

        Returns:
            TXResponse: Transaction response.
        """

        deserializer = deserializer if deserializer is not None else TXResponse.deserializer

        return await self._rpc.call("state.Delegate", (del_addr, str(amount), config), deserializer)

    async def grant_fee(self, grantee: str, amount: int, deserializer: Callable | None = None,
                        **config: Unpack[TxConfig]) -> TXResponse:
        """ Grants a fee allowance to the specified grantee.

        Args:
            grantee (str): Address of the grantee.
            amount (int): Amount of fee allowance.
            deserializer (Callable | None): Custom deserializer. Defaults to :meth:`~celestia.types.state.TXResponse.deserializer`.
            **config(TxConfig): Additional transaction configurations.

        Returns:
            TXResponse: Response of the grant fee transaction.
        """

        deserializer = deserializer if deserializer is not None else TXResponse.deserializer

        return await self._rpc.call("state.GrantFee", (grantee, str(amount), config), deserializer)

    async def query_delegation(self, val_addr: str, deserializer: Callable | None = None) -> QueryDelegationResponse:
        """ Retrieves the delegation information between a delegator and a validator.

        Args:
            val_addr (str): Validator address.
            deserializer (Callable | None): Custom deserializer. Defaults to :meth:`~celestia.types.state.QueryDelegationResponse.deserializer`.

        Returns:
            QueryDelegationResponse: Delegation information.
        """

        deserializer = deserializer if deserializer is not None else QueryDelegationResponse.deserializer

        return await self._rpc.call("state.QueryDelegation", (val_addr,), deserializer)

    async def query_redelegations(self, src_val_addr: str, dst_val_addr: str,
                                  deserializer: Callable | None = None) -> QueryRedelegationResponse:
        """ Retrieves the status of the redelegations between a delegator and a validator.

        Args:
            src_val_addr (str): Source validator address.
            dst_val_addr (str): Destination validator address.
            deserializer (Callable | None): Custom deserializer. Defaults to :meth:`~celestia.types.state.QueryRedelegationResponse.deserializer`.

        Returns:
            QueryRedelegationResponse: Redelegation details.
        """

        deserializer = deserializer if deserializer is not None else QueryRedelegationResponse.deserializer

        return await self._rpc.call("state.QueryRedelegations", (src_val_addr, dst_val_addr), deserializer)

    async def query_unbonding(self, val_addr: str,
                              deserializer: Callable | None = None) -> QueryUnbondingDelegationResponse:
        """ Retrieves the unbonding status between a delegator and a validator.

        Args:
            val_addr (str): Validator address.
            deserializer (Callable | None): Custom deserializer. Defaults to :meth:`~celestia.types.state.QueryUnbondingDelegationResponse.deserializer`.

        Returns:
           QueryUnbondingDelegationResponse: Unbonding status.
        """

        deserializer = deserializer if deserializer is not None else QueryUnbondingDelegationResponse.deserializer

        return await self._rpc.call("state.QueryUnbonding", (val_addr,), deserializer)

    async def revoke_grant_fee(self, grantee: str, deserializer: Callable | None = None,
                               **config: Unpack[TxConfig]) -> TXResponse:
        """ Revokes a previously granted fee allowance.

        Args:
            grantee (str): Address of the grantee whose allowance is being revoked.
            deserializer (Callable | None): Custom deserializer. Defaults to :meth:`~celestia.types.state.TXResponse.deserializer`.
            **config(TxConfig): Additional transaction configurations.

        Returns:
            TXResponse: Response of the revoke grant fee transaction.
        """

        deserializer = deserializer if deserializer is not None else TXResponse.deserializer

        return await self._rpc.call("state.RevokeGrantFee", (grantee, config), deserializer)

    async def submit_pay_for_blob(self, blob: Blob, *blobs: Blob, **config: Unpack[TxConfig]) -> int:
        """ Builds, signs and submits a PayForBlob transaction.

        Args:
            blob (Blob): The first blob to be included in the transaction.
            *blobs (Blob): Additional blobs.
            **config(TxConfig): Additional transaction configurations.

        Returns:
            int: Transaction ID of the submitted PayForBlob transaction.
        """
        blobs = tuple(types.normalize_blob(blob) if blob.commitment is None else blob for blob in (blob, *blobs))
        return await self._rpc.call("state.SubmitPayForBlob", (blobs, config))

    async def transfer(self, to: str, amount: int, deserializer: Callable | None = None,
                       **config: Unpack[TxConfig]) -> TXResponse:
        """ Sends the given amount of coins from default wallet of the node to the given account address.

        Args:
            to (str): Recipient address.
            amount (int): Amount to transfer.
            deserializer (Callable | None): Custom deserializer. Defaults to :meth:`~celestia.types.state.TXResponse.deserializer`.
            **config(TxConfig): Additional transaction configuration.

        Returns:
            TXResponse: Transaction response.
        """

        deserializer = deserializer if deserializer is not None else TXResponse.deserializer

        return await self._rpc.call("state.Transfer", (to, str(amount), config), deserializer)

    async def undelegate(self, del_addr: str, amount: int, deserializer: Callable | None = None,
                         **config: Unpack[TxConfig]) -> TXResponse:
        """ Undelegates a user's delegated tokens, unbonding them from the current validator.

        Args:
            del_addr (str): Delegator address.
            amount (int): Amount to undelegate.
            deserializer (Callable | None): Custom deserializer. Defaults to :meth:`~celestia.types.state.TXResponse.deserializer`.
            **config(TxConfig): Additional transaction configurations.

        Returns:
            TXResponse: Response of the undelegation transaction.
        """

        deserializer = deserializer if deserializer is not None else TXResponse.deserializer

        return await self._rpc.call("state.Undelegate", (del_addr, str(amount), config), deserializer)



from collections.abc import AsyncIterator

from ._RPC import Wrapper


class FraudClient(Wrapper):
    """ Client for interacting with Celestia's Fraud API."""

    async def get(self, proof_type: str) -> list[dict[str, str]]:
        """ Fetches fraud proofs from the disk by its type.

        Args:
            proof_type (str): The type of fraud proof to retrieve.

        Returns:
            list[dict[str, str]]: A list of fraud proofs.
        """
        return await self._rpc.call("fraud.Get", (proof_type,))

    async def subscribe(self, proof_type: str) -> AsyncIterator[dict[str, str]]:
        """ Allows to subscribe on a Proof pub sub topic by its type.

        Args:
            proof_type (str): The type of fraud proof to subscribe to.

        Yields:
            dict[str, str]: A dictionary containing fraud proof data.
        """
        return self._rpc.iter("fraud.Subscribe", (proof_type,))



import typing as t

from celestia.types.das import SamplingStats
from ._RPC import Wrapper


class DasClient(Wrapper):
    """ Client for interacting with Celestia's Das API."""

    async def sampling_stats(self, deserializer: t.Callable | None = None) -> SamplingStats:
        """ Returns the current statistics over the DA sampling process.

        Args:
            deserializer (Callable | None): Custom deserializer. Defaults to :meth:`~celestia.types.das.SamplingStats.deserializer`.

        Returns:
            SamplingStats: The current sampling statistics.
        """

        deserializer = deserializer if deserializer is not None else SamplingStats.deserializer

        return await self._rpc.call("das.SamplingStats", (), deserializer)

    async def wait_catch_up(self) -> None:
        """ Blocks until DASer finishes catching up to the network head.

        Returns:
            None
        """
        return await self._rpc.call("das.WaitCatchUp")



from dataclasses import dataclass

from celestia._celestia import types as ext  # noqa

from celestia.types.common_types import Base64


@dataclass
class ConsensusVersion:
    """ Represents the version information for the consensus.

    Attributes:
        block (str): The block version.
        app (str): The application version.
    """
    block: str
    app: str


@dataclass
class Parts:
    """ Represents the parts of the block.

    Attributes:
        total (int): The total number of parts.
        hash (str): The hash of the parts.
    """
    total: int
    hash: str


@dataclass
class BlockId:
    """ Represents a block identifier, which includes a hash and parts.

    Attributes:
        hash (str): The hash of the block.
        parts (Parts): The parts information of the block.
    """
    hash: str
    parts: Parts

    def __init__(self, hash, parts):
        self.hash = hash
        self.parts = Parts(**parts)


@dataclass
class Header:
    """ Represents the header information for the block.

    Attributes:
        version (ConsensusVersion): The consensus version.
        chain_id (str): The chain identifier.
        height (str): The height of the block.
        time (str): The time the block was created.
        last_block_id (BlockId): The identifier of the last block.
        last_commit_hash (str): The hash of the last commit.
        data_hash (str): The hash of the block data.
        validators_hash (str): The hash of the validators.
        next_validators_hash (str): The hash of the next validators.
        consensus_hash (str): The consensus hash.
        app_hash (str): The application hash.
        last_results_hash (str): The hash of the last results.
        evidence_hash (str): The hash of the evidence.
        proposer_address (str): The address of the proposer.
    """
    version: ConsensusVersion
    chain_id: str
    height: str
    time: str
    last_block_id: BlockId
    last_commit_hash: str
    data_hash: str
    validators_hash: str
    next_validators_hash: str
    consensus_hash: str
    app_hash: str
    last_results_hash: str
    evidence_hash: str
    proposer_address: str

    def __init__(self, version, chain_id, height, time, last_block_id, last_commit_hash, data_hash, validators_hash,
                 next_validators_hash, consensus_hash, app_hash, last_results_hash, evidence_hash, proposer_address):
        self.version = ConsensusVersion(**version)
        self.chain_id = chain_id
        self.height = height
        self.time = time
        self.last_block_id = BlockId(**last_block_id)
        self.last_commit_hash = last_commit_hash
        self.data_hash = data_hash
        self.validators_hash = validators_hash
        self.next_validators_hash = next_validators_hash
        self.consensus_hash = consensus_hash
        self.app_hash = app_hash
        self.last_results_hash = last_results_hash
        self.evidence_hash = evidence_hash
        self.proposer_address = proposer_address


@dataclass
class PubKey:
    """ Represents a public key used for validating a transaction.

    Attributes:
        type (str): The type of public key.
        value (Base64): The base64 encoded public key.
    """
    type: str
    value: Base64


@dataclass
class Validator:
    """ Represents a validator in the consensus system.

    Attributes:
        address (str): The address of the validator.
        pub_key (PubKey): The public key of the validator.
        voting_power (str): The voting power of the validator.
        proposer_priority (str): The proposer priority of the validator.
    """
    address: str
    pub_key: PubKey
    voting_power: str
    proposer_priority: str

    def __init__(self, address, pub_key, voting_power, proposer_priority):
        self.address = address
        self.pub_key = PubKey(**pub_key)
        self.voting_power = voting_power
        self.proposer_priority = proposer_priority


@dataclass
class ValidatorSet:
    """ Represents a set of validators and the proposer.

    Attributes:
        validators (tuple[Validator, ...]): The list of validators.
        proposer (Validator): The proposer of the block.
    """
    validators: tuple[Validator, ...]
    proposer: Validator

    def __init__(self, validators, proposer):
        self.validators = tuple(Validator(**validator) for validator in validators)
        self.proposer = Validator(**proposer)


@dataclass
class Signature:
    """ Represents a signature for a commit block.

    Attributes:
        block_id_flag (int): The block ID flag.
        validator_address (str): The address of the validator signing the block.
        timestamp (str): The timestamp of the signature.
        signature (Base64): The base64 encoded signature.
    """
    block_id_flag: int
    validator_address: str
    timestamp: str
    signature: Base64

    def __init__(self, block_id_flag, validator_address, timestamp, signature):
        self.block_id_flag = block_id_flag
        self.validator_address = validator_address
        self.timestamp = timestamp
        self.signature = signature


@dataclass
class Commit:
    """ Represents a commit for a block, including signatures.

    Attributes:
        height (int): The block height.
        round (int): The block round.
        block_id (BlockId): The ID of the block.
        signatures (tuple[Signature, ...]): The signatures of the validators.
    """
    height: int
    round: int
    block_id: BlockId
    signatures: tuple[Signature, ...]

    def __init__(self, height, round, block_id, signatures):
        self.height = height
        self.round = round
        self.block_id = BlockId(**block_id)
        self.signatures = tuple(Signature(**signature) for signature in signatures)


@dataclass
class Dah:
    """ Represents the data availability header.

    Attributes:
        row_roots (tuple[Base64, ...]): The row roots.
        column_roots (tuple[Base64, ...]): The column roots.
    """
    row_roots: tuple[Base64, ...]
    column_roots: tuple[Base64, ...]

    def __init__(self, row_roots, column_roots):
        self.row_roots = tuple(row_root for row_root in row_roots)
        self.column_roots = tuple(column_root for column_root in column_roots)


@dataclass
class ExtendedHeader:
    """ Represents an extended header containing header, validator set, commit, and DAH.

    Attributes:
        header (Header): The block header.
        validator_set (ValidatorSet): The validator set.
        commit (Commit): The commit information.
        dah (Dah): The data availability header.
    """
    header: Header
    validator_set: ValidatorSet
    commit: Commit
    dah: Dah

    def __init__(self, header, validator_set, dah, commit):
        self.header = Header(**header)
        self.validator_set = ValidatorSet(**validator_set)
        self.commit = Commit(**commit)
        self.dah = Dah(**dah)

    @staticmethod
    def deserializer(result: dict) -> 'ExtendedHeader':
        """ Deserializes the provided result into a `ExtendedHeader` object.

        Args:
            result (dict): The dictionary representation of a ExtendedHeader.

        Returns:
            ExtendedHeader: A deserialized ExtendedHeader object.
        """
        if result is not None:
            return ExtendedHeader(**result)


@dataclass
class State:
    """ Represents a state for the block range.

    Attributes:
        id (int): The ID of the state.
        height (int): The height of the block.
        from_height (int): The starting height for the range.
        to_height (int): The ending height for the range.
        from_hash (str): The hash at the start of the range.
        to_hash (str): The hash at the end of the range.
        start (str): The start time of the state.
        end (str): The end time of the state.
    """
    id: int
    height: int
    from_height: int
    to_height: int
    from_hash: str
    to_hash: str
    start: str
    end: str

    def __init__(self, id, height, from_height, to_height, from_hash, to_hash, start, end):
        self.id = id
        self.height = height
        self.from_height = from_height
        self.to_height = to_height
        self.from_hash = from_hash
        self.to_hash = to_hash
        self.start = start
        self.end = end

    @staticmethod
    def deserializer(result: dict) -> 'State':
        """ Deserializes the provided result into a `State` object.

        Args:
            result (dict): The dictionary representation of a State.

        Returns:
            State: A deserialized State object.
        """
        if result is not None:
            return State(**result)



import subprocess
from dataclasses import dataclass
from io import StringIO


@dataclass
class Container:
    id: str
    name: str
    image: str
    port: dict[str, int]

    def __init__(self, id: str, name: str, image: str, ports: str):
        self.id = id
        self.name = name
        self.image = image
        ports = [s.strip().split('->') for s in ports.split(',')]
        self.port = dict((p[1], int(p[0].split(':')[-1])) for p in ports if len(p) == 2)


class Containers(list[Container]):

    def __init__(self, prefix: str):
        result = list()
        self._prefix = prefix
        proc = subprocess.run(['docker', 'ps'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        f = StringIO(proc.stdout.decode('utf8'))
        ready = f.readline()
        if ready:
            ready = ready.replace("CONTAINER ID", "CONTAINER_ID")
            b = [ready.index(word) for word in ready.split()] + [1024]
            while True:
                if line := f.readline():
                    line.split()
                    c = [line[b[N]:b[N + 1]].strip() for N in range(len(b) - 1)]
                    if c[6].startswith(self._prefix):
                        result.append(Container(c[0], c[6], c[1], c[5]))
                else:
                    break
        super().__init__(result)

    def get_by_id(self, id: str) -> Container | None:
        if result := [container for container in self if container.id == id]:
            return result[0]

    def get_by_name(self, name: str) -> tuple[Container, ...]:
        return tuple(sorted([container for container in self
                             if container.name.startswith('-'.join([self._prefix, name, '']))],
                            key=lambda container: container.name))

    def get_by_name_first(self, name: str) -> Container | None:
        if result := self.get_by_name(name):
            return result[0]



import pytest

from celestia.node_api import Client
from celestia.types.common_types import Blob
from celestia.types.share import SampleCoords


@pytest.mark.asyncio
async def test_share(node_provider):
    bridge, auth_token = await node_provider('bridge-0')
    client = Client(port=bridge.port['26658/tcp'])
    async with client.connect(auth_token) as api:
        result = await api.blob.submit(
            Blob(b'abc', b'0123456789'),
            Blob(b'abc', b'QWERTYUIOP'),
            Blob(b'xyz', b'ASDFGHJKL'),
        )

        eds = await api.share.get_eds(result.height)
        gnd = await api.share.get_namespace_data(result.height, b'abc')
        range_data = await api.share.get_range(result.height, 1, 2)
        samples = await api.share.get_samples((await api.header.get_by_height(result.height)),
                                              [SampleCoords(row=0, col=1)])
        coords_data = await api.share.get_share(result.height, 0, 1)
        assert coords_data == samples[0] == range_data.proof.data[0] == eds.data_square[1] == gnd[0].shares[0]
        await api.share.get_available(result.height)



import pytest

from celestia.node_api import Client


@pytest.mark.asyncio
async def test_das(node_provider):
    light, auth_token = await node_provider('light-0')
    client = Client(port=light.port['26658/tcp'])

    async with client.connect(auth_token) as api:
        await api.das.wait_catch_up()
        assert (await api.das.sampling_stats()).catch_up_done



import asyncio

import pytest

from celestia.node_api import Client
from celestia.types.common_types import Namespace, Blob


@pytest.mark.asyncio
async def test_send_blobs(node_provider):
    bridge, auth_token = await node_provider('bridge-0')
    client = Client(port=bridge.port['26658/tcp'])
    async with client.connect(auth_token) as api:
        blobs = await api.blob.get_all(1, b'abc')
        assert blobs is None

        result = await api.blob.submit(
            Blob(b'abc', b'0123456789'),
            Blob(b'abc', b'QWERTYUIOP'),
            Blob(b'xyz', b'ASDFGHJKL'),
        )

        assert len(result.commitments) == 3

        blobs = await api.blob.get_all(result.height, b'abc')
        assert len(blobs) == 2
        assert blobs[0].data == b'0123456789'
        assert blobs[0].commitment == result.commitments[0]

        assert blobs[1].data == b'QWERTYUIOP'
        assert blobs[1].commitment == result.commitments[1]

        blobs = await api.blob.get_all(result.height, b'xyz')
        assert len(blobs) == 1
        assert blobs[0].data == b'ASDFGHJKL'
        assert blobs[0].commitment == result.commitments[2]

        blob = await api.blob.get(1, b'abc', b'ASDFGHJKL')
        assert blob is None

        blob = await api.blob.get(result.height, b'abc', result.commitments[1])
        assert len(blobs) == 1
        assert blob.data == b'QWERTYUIOP'

        proof = await api.blob.get_proof(1, b'abc', b'ASDFGHJKL')
        assert proof is None

        proof = await api.blob.get_proof(result.height, b'abc', result.commitments[1])
        assert proof is not None

        included = await api.blob.included(result.height, b'xyz', proof, result.commitments[1])
        assert not included

        included = await api.blob.included(result.height, b'abc', proof, result.commitments[1])
        assert included

        com_proof = await api.blob.get_commitment_proof(result.height, b'abc', result.commitments[1])
        assert com_proof is not None

        com_proof = await api.blob.get_commitment_proof(result.height, b'abc', result.commitments[1])
        assert com_proof is not None


@pytest.mark.asyncio
async def test_blob_exceptions(node_provider):
    bridge, auth_token = await node_provider('bridge-0')
    client = Client(port=bridge.port['26658/tcp'])
    async with client.connect(auth_token) as api:
        with pytest.raises(ValueError):
            await api.blob.get_all(18446744073709551615, b'abc')
        with pytest.raises(ValueError):
            await api.blob.get_all(0, b'abc')
        with pytest.raises(ValueError):
            await api.blob.get_all(-1, b'abc')

        with pytest.raises(ValueError):
            await api.blob.get(18446744073709551615, b'abc', b'ASDFGHJKL')
        with pytest.raises(ValueError):
            await api.blob.get(0, b'abc', b'ASDFGHJKL')
        with pytest.raises(ValueError):
            await api.blob.get(-1, b'abc', b'ASDFGHJKL')

        with pytest.raises(ValueError):
            await api.blob.get_proof(18446744073709551615, b'abc', b'ASDFGHJKL')
        with pytest.raises(ValueError):
            await api.blob.get_proof(0, b'abc', b'ASDFGHJKL')
        with pytest.raises(ValueError):
            await api.blob.get_proof(-1, b'abc', b'ASDFGHJKL')

        with pytest.raises(TypeError):
            Blob(b'abc', None)
        with pytest.raises(ValueError):
            Blob(123, b'ASDFGHJKL')


@pytest.mark.asyncio
async def test_blob_subscribe(node_provider):
    bridge, auth_token = await node_provider('bridge-0')
    client = Client(port=bridge.port['26658/tcp'])

    result = []

    async def submitter(api):
        for i in range(5):
            if i == 2:
                await api.blob.submit(Blob(b'abc', f'QWERTY{i}'.encode()))
            else:
                await api.blob.submit(Blob(b'qwe', f'QWERTY{i}'.encode()))

    async with client.connect(auth_token) as api:
        submitter_tack = asyncio.create_task(submitter(api))
        try:
            async with asyncio.timeout(30):
                async for blob in api.blob.subscribe(b'qwe'):
                    result.append(blob)
                    if len(result) == 4:
                        break
        finally:
            submitter_tack.cancel()

    assert len(result) == 4
    assert tuple(item.blobs[0].namespace for item in result) == (Namespace(b'qwe'),) * 4



import pytest
from celestia._celestia import types  # noqa

from celestia.types.common_types import Namespace


@pytest.mark.asyncio
async def test_namespace():
    namespace = Namespace(b'Alesh')
    assert namespace == b'\x00' * 24 + b'Alesh'
    assert str(namespace) == 'AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAQWxlc2g='
    assert namespace == Namespace('AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAQWxlc2g=')
    with pytest.raises(ValueError):
        Namespace(b'111111111111111111111111111111111')
    with pytest.raises(ValueError):
        Namespace(b'1111111111111111111111111111111111111111111111111111111111')


@pytest.mark.asyncio
async def test_blob():
    blob = types.normalize_blob(b'Alesh', b'0123456789ABCDEF')
    assert blob == {
        'namespace': b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00Alesh',
        'commitment': b'\x88e\rh\xc1\x02\xbd\xfc\xbcc\xa3\xcc\x10\n5\xdf\xcbCh\xa3m\x04\xe1\xeds(\xdf}j>\xab/',
        'data': b'0123456789ABCDEF', 'index': None, 'share_version': 0
    }



import asyncio
import typing as t
from asyncio import CancelledError
from contextlib import AbstractAsyncContextManager, asynccontextmanager

from websockets.asyncio.client import connect, ClientConnection

from ._RPC import RPC
from .blob import BlobClient
from .das import DasClient
from .fraud import FraudClient
from .header import HeaderClient
from .p2p import P2PClient
from .share import ShareClient
from .state import StateClient


class Client:
    """ Celestia Node API client
    """

    def __init__(self, auth_token: str = None,
                 host: str = 'localhost', port: int = 26658,
                 response_timeout: int = 180, **options: t.Any):
        self.__options = dict(options, host=host, port=port,
                              auth_token=auth_token, response_timeout=response_timeout)

    @property
    def options(self):
        """ Client create options """
        return self.__options

    @property
    def url(self):
        """ Connection URL """
        return f'ws://{self.options["host"]}:{self.options["port"]}'

    class NodeAPI:
        """ Celestia node API
        """

        def __init__(self, rpc: RPC):
            self._rpc = rpc

        @property
        def state(self):
            return StateClient(self._rpc)

        @property
        def blob(self):
            return BlobClient(self._rpc)

        @property
        def header(self):
            return HeaderClient(self._rpc)

        @property
        def p2p(self):
            return P2PClient(self._rpc)

        @property
        def das(self):
            return DasClient(self._rpc)

        @property
        def fraud(self):
            return FraudClient(self._rpc)

        @property
        def share(self):
            return ShareClient(self._rpc)

    def connect(self, auth_token: str = None, **options: t.Any) -> AbstractAsyncContextManager[NodeAPI]:
        """ Creates and return connection context manager. """
        headers = []
        options = dict(self.options, **options)
        response_timeout = options['response_timeout']
        auth_token = auth_token or options['auth_token']
        if auth_token is not None:
            headers.append(('Authorization', f'Bearer {auth_token}'))

        async def listener(connection: ClientConnection, receiver: t.Callable[[str | bytes], None]):
            try:
                async for message in connection:
                    receiver(message)
            except CancelledError:
                pass

        @asynccontextmanager
        async def connect_context():
            rpc = RPC(response_timeout)
            async with connect(self.url, additional_headers=headers) as connection:
                async with rpc.connect(connection) as receiver:
                    self._listener_task = asyncio.create_task(listener(connection, receiver))
                    yield self.NodeAPI(rpc)
                    self._listener_task.cancel()

        return connect_context()


NodeAPI = Client.NodeAPI
